import logging
from typing import Callable, Optional


class Response:
    encoding = None

    def __init__(self, url: str, method: str, *, status: int = 0, encoding: Optional[str] = None,
                 headers: Optional[dict] = None, cb_text: Callable = None, cb_json: Callable = None):
        self.url = url
        self.method = method
        self.encoding = encoding
        self.headers = headers
        self.cb_text = cb_text
        self.cb_json = cb_json
        self.status = status

    async def text(self, encoding: Optional[str] = None, errors: str = 'strict'):
        """aiohttp response text"""
        text = await self.cb_text(encoding, errors=errors)
        logging.info(f'response [{self.method}] {self.url}')
        return text



"""
test the integer field
"""
import asyncio

from gerridae import Item, RegexField, Spider
from gerridae.log import get_logger

logger = get_logger(__name__)


class TestRegexField(object):
    def test_regex_filed(self):
        class TestItem(Item):
            command = RegexField(pattern='id="pip-command">(.*?)</span>', many=False)

        class TestSpider(Spider):
            start_urls = "https://pypi.org/project/gerridae/"

            async def parse(self, resp):
                text = await resp.text(encoding='utf-8')
                result = TestItem.get_item(html=text)
                return result

        results = asyncio.run(TestSpider.start())
        print(results)
        assert "pip install gerridae" == results[0].command

    def test_regex_field_with_many(self):
        class TestItem(Item):
            command = RegexField(pattern='id="pip-command">(.*?)</span>', many=True)

        class TestSpider(Spider):
            start_urls = 'https://pypi.org/project/gerridae/'

            async def parse(self, resp):
                text = await resp.text(encoding='utf-8')
                result = TestItem.get_item(html=text)
                return result

        results = asyncio.run(TestSpider.start())
        assert ['pip install gerridae'] == results[0].command



import asyncio
import logging
from typing import Optional

from .response import Response


class Request:
    status = 0

    def __init__(self, url: str, method: str = 'get', *, delay: int = 0, encoding: str = None, request_session=None,
                 headers: Optional[dict] = None, **aiohttp_kwargs):
        """request init"""
        self.url = url
        self.method = method
        self.request_session = request_session
        self.delay = delay
        self.encoding = encoding
        self.headers = headers
        self.aiohttp_kwargs = aiohttp_kwargs

    async def fetch(self):
        """fetch request"""
        if self.delay:
            await asyncio.sleep(self.delay)

        resp = await self._make_response()
        response = Response(
            url=self.url,
            method=self.method,
            encoding=self.encoding,
            status=resp.status,
            headers=self.headers,
            cb_text=resp.text,
            cb_json=resp.json
        )
        return response

    async def _make_response(self):
        logging.debug(f'request [{self.method}]  {self.url}')
        resp = await self.request_session.get(self.url, headers=self.headers, **self.aiohttp_kwargs)
        return resp

    async def retry(self, retry_time: int = 0):
        logging.info(f'Retry [{self.method}] {self.url}')
        return await self.fetch()

    def __repr__(self):
        return f'<Request{self.url} <status> {self.status}'



from gerridae.field import BaseField
from lxml import etree


class ItemMeta(type):
    def __new__(cls, name, bases, attrs):
        __fields = {
            field_name: attrs.pop(field_name)
            for field_name, obj in list(attrs.items())
            if isinstance(obj, BaseField)
        }
        attrs["__fields"] = __fields

        return type.__new__(cls, name, bases, attrs)


class Item(metaclass=ItemMeta):
    def __init__(self):
        self.result = {}

    def __repr__(self):
        return f"<{self.__class__.__name__}> {self.result}"

    @staticmethod
    def _get_html(html, **kwargs):
        """获取 etree.HTML obj"""
        if html:
            return etree.HTML(html)
        raise ValueError("html or url is expected")

    @classmethod
    def _parser_html(cls, *, html_etree):
        if html_etree is None:
            return ValueError("html_etree is expected")
        item_ins = cls()

        fields = getattr(item_ins, "__fields", {})
        for field_name, field_value in fields.items():
            clean_method = getattr(item_ins, f"clean_{field_name}", None)
            value = field_value.extract(html_etree)

            # check the many to get the current
            many = getattr(field_value, "many", False)
            if not many:
                value = value if len(value) == 0 or isinstance(value, str) else value[0]

            if clean_method and callable(clean_method):
                assert value, f"{field_name} value is {value}"
                value = clean_method(value)

            # default value for field
            if not value:
                default_value = getattr(field_value, "default", False)
                value = default_value

            # setattr for Item class
            setattr(item_ins, field_name, value)
            item_ins.result[field_name] = value
        return item_ins

    @classmethod
    def get_item(cls, html=None):
        if html:
            html_etree = cls._get_html(html)
            return cls._parser_html(html_etree=html_etree)
        raise ValueError("html or url is excepted")

    @classmethod
    def get_items(cls, html=None, url=None):
        pass



from setuptools import find_packages, setup

__version__ = "0.3.3"


def long_description():
    with open("README.md", "r") as f:
        long_desc = f.read()
    return long_desc


setup(
    name="gerridae",
    version=__version__,
    author="buglan",
    packages=find_packages(),
    description="a crawl framework just for fun",
    long_description=long_description(),
    long_description_content_type="text/markdown",
    author_email="dev.buglan@gmail.com",
    python_requires=">=3.6",
    install_requires=["cssselect", "lxml", "aiohttp"],
    url="https://github.com/BUGLAN/gerridae",
    setup_requires=["pytest-runner"],
    tests_require=["pytest"],
)



import asyncio
from datetime import datetime

import aiohttp

from .log import get_logger
from .request import Request, Response

logger = get_logger(__name__)

HEADER = {
    "User-Agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_14_5) AppleWebKit/537.36 (KH"
                  "TML, like Gecko) Chrome/75.0.3770.100 Safari/537.36"
}


class Spider:
    def __init__(self):
        pass

    start_urls = []
    encoding = None
    name = 'Spider'
    sender = 0
    returned = 0

    @staticmethod
    def is_crawl(url):
        return False

    def parse(self, rep):
        raise NotImplementedError

    @classmethod
    async def start(cls):
        logger.info(f'{cls.name} Start Crawl')
        start_time, spider_ins = datetime.now(), cls()

        if isinstance(spider_ins.start_urls, str):
            spider_ins.start_urls = [spider_ins.start_urls]

        async with aiohttp.ClientSession() as session:
            loop = asyncio.get_running_loop()
            loop.set_debug(True)
            logger.info(f'start create task except create {len(spider_ins.start_urls)} task')
            tasks = [asyncio.create_task(spider_ins.request(url, session=session)) for url in spider_ins.start_urls if
                     not spider_ins.is_crawl(url)]
            logger.info(f'already create  {len(tasks)} tasks')
            finished, unfinished = await asyncio.wait(tasks)
            logger.info(f'fail task num: {len(unfinished)}')
            data = [r.result() for r in finished]
            logger.info(f"{cls.name} cost time {datetime.now() - start_time}")
            return data

    async def request(self, url, *, session=None) -> Response:
        resp = await Request(url, request_session=session).fetch()
        return await self.parse(resp)



import html
import re

from lxml import etree


class BaseField:
    """base field"""

    def __init__(self, many=False, default=None):
        self.default = default
        self.many = many

    def extract(self, *args, **kwargs):
        raise NotImplementedError("extract is not implemented.")


class _LXMLElementField(BaseField):
    """usage TextField Attribute """

    def __init__(self, css_select=None, many=False, default=None):
        super(_LXMLElementField, self).__init__(many, default)
        self.css_select = css_select
        self.many = many

    def _get_elements(self, *, html_etree):
        if self.css_select:
            elements = html_etree.cssselect(self.css_select)
        else:
            raise ValueError(f"{self.__class__.__name__} field css_select is expected")
        return elements

    def _parser_element(self, element):
        raise NotImplementedError

    def extract(self, html_etree):
        elements = self._get_elements(html_etree=html_etree)
        if elements:
            result = [self._parser_element(element) for element in elements]
        else:
            result = []
        return result


class TextField(_LXMLElementField):
    """text filed"""

    def __init__(self, css_select=None, many=False, default=None):
        super(TextField, self).__init__(css_select, many, default)

    def _parser_element(self, element):
        strings = [node.strip() for node in element.itertext()]
        string = "".join(strings)
        return string if string else self.default


class AttrField(_LXMLElementField):
    """attribute field"""

    def __init__(self, css_select=None, attr=None, many=False, default=None):
        super(AttrField, self).__init__(css_select, many, default)
        assert attr, "excepted attr"
        self.attr = attr

    def _parser_element(self, element):
        return element.get(self.attr, self.default)


class RegexField(BaseField):
    def __init__(self, pattern=None, re_flags=0, many=False, default=None, encoding='utf-8'):
        super(RegexField, self).__init__(many, default)
        self.encoding = encoding
        self.pattern = pattern
        self._re_object = re.compile(pattern, flags=re_flags)

    def _parse_match(self, match):
        if not match:
            if self.default is None:
                raise ValueError(
                    f"""Extract `{self.pattern}` error, "please check selector or set parameter named `default`""")
            return self.default

        string = match.group()
        groups = match.groups()
        group_dict = match.groupdict()
        if group_dict:
            return group_dict
        if groups:
            return groups[0] if len(groups) == 1 else groups
        return string

    def extract(self, html_etree):
        html_data = etree.tostring(html_etree).decode(encoding=self.encoding)
        html_data = self.decode_html(html_data)

        if self.many:
            matches = self._re_object.finditer(html_data)
            return [self._parse_match(match) for match in matches]
        else:
            match = self._re_object.search(html_data)
            return self._parse_match(match)

    @staticmethod
    def decode_html(html_data):
        s = html.unescape(html_data)
        return s



"""
test the integer field
"""
import asyncio

from gerridae import Item, Spider, TextField, RegexField
from gerridae.log import get_logger

logger = get_logger(__name__)


class TestBaseField:
    def test_spider(self):
        class TestItem(Item):
            command = TextField(css_select="#pip-command")

        class TestSpider(Spider):
            start_urls = "https://pypi.org/project/gerridae/"

            async def parse(self, resp):
                text = await resp.text(encoding='utf-8')
                result = TestItem.get_item(html=text)
                return result

        results = asyncio.run(TestSpider.start())
        assert "pip install gerridae" == results[0].command

    def test_spider_with_many(self):
        class TestItem(Item):
            command = TextField(css_select="#pip-command", many=True)

        class TestSpider(Spider):
            start_urls = "https://pypi.org/project/gerridae/"

            async def parse(self, resp):
                text = await resp.text(encoding='utf-8')
                result = TestItem.get_item(html=text)
                return result

        results = asyncio.run(TestSpider.start())
        assert ["pip install gerridae"] == results[0].command

    def test_spider_with_multi_urls(self):
        class TestItem(Item):
            command = TextField(css_select="#pip-command", many=True)

        class TestSpider(Spider):
            start_urls = ["https://pypi.org/project/gerridae/", "https://pypi.org/project/gerridae/"]

            async def parse(self, resp):
                text = await resp.text(encoding='utf-8')
                result = TestItem.get_item(html=text)
                setattr(result, 'base_url', resp.url)
                return result

        results = asyncio.run(TestSpider.start())
        assert ["pip install gerridae"] == results[0].command
        assert ["pip install gerridae"] == results[1].command
        assert "https://pypi.org/project/gerridae/" == results[0].base_url

    def test_spider_regex_field(self):
        class TestItem(Item):
            content = RegexField(pattern=r'<div id="content">(.*?)</div>')

        class TestSpider(Spider):
            start_urls = 'https://www.230book.com/book/5483/2065394.html'

            async def parse(self, resp):
                text = await resp.text(encoding='gbk')
                result = TestItem.get_item(html=text)
                return result

        results = asyncio.run(TestSpider.start())
        assert len(results) == 1
        assert len(results[0].content) > 0
